@@ -1,5 +1,5 @@
 /*
- * $Id: pinboard.c,v 1.167.2.1 2003/07/23 15:45:10 tal197 Exp $
+ * $Id: pinboard.c,v 1.174 2004/03/06 18:50:48 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
  * Copyright (C) 2003, the ROX-Filer team.
@@ -146,6 +146,8 @@ static Option o_label_font, o_pinboard_shadow_colour;
 static Option o_pinboard_shadow_labels;
 static Option o_blackbox_hack;
 
+static Option o_top_margin, o_bottom_margin;
+
 /* Static prototypes */
 static GType pin_icon_get_type(void);
 static void set_size_and_style(PinIcon *pi);
@@ -210,7 +212,6 @@ static void create_pinboard_window(Pinboard *pinboard);
 static void reload_backdrop(Pinboard *pinboard,
 			    const gchar *backdrop,
 			    BackdropStyle backdrop_style);
-static void set_backdrop(const gchar *path, BackdropStyle style);
 static void pinboard_reshape_icon(Icon *icon);
 static gint draw_wink(GtkWidget *widget, GdkEventExpose *event, PinIcon *pi);
 static void abandon_backdrop_app(Pinboard *pinboard);
@@ -251,6 +252,9 @@ void pinboard_init(void)
 
 	option_add_int(&o_blackbox_hack, "blackbox_hack", FALSE);
 
+	option_add_int(&o_top_margin, "pinboard_top_margin", 0);
+	option_add_int(&o_bottom_margin, "pinboard_bottom_margin", 0);
+
 	option_add_notify(pinboard_check_options);
 
 	gdk_color_parse(o_pinboard_fg_colour.value, &pin_text_fg_col);
@@ -554,7 +558,7 @@ void pinboard_set_backdrop_app(const gchar *app)
 		g_object_unref(ai);
 
 	if (can_set)
-		set_backdrop(app, BACKDROP_PROGRAM);
+		pinboard_set_backdrop(app, BACKDROP_PROGRAM);
 	else
 		delayed_error(_("You can only set the backdrop to an image "
 				"or to a program which knows how to "
@@ -565,7 +569,7 @@ void pinboard_set_backdrop_app(const gchar *app)
 }
 
 /* Open a dialog box allowing the user to set the backdrop */
-void pinboard_set_backdrop(void)
+void pinboard_set_backdrop_box(void)
 {
 	GtkWidget *dialog, *frame, *label, *hbox;
 	GtkBox *vbox;
@@ -660,6 +664,71 @@ void draw_label_shadow(WrappedLabel *wl, GdkRegion *region)
 	gdk_gc_set_clip_region(current_pinboard->shadow_gc, NULL);
 }
 
+/* Set and save (path, style) as the new backdrop.
+ * If style is BACKDROP_PROGRAM, the program is run to get the backdrop.
+ * Otherwise, the image is displayed now.
+ */
+void pinboard_set_backdrop(const gchar *path, BackdropStyle style)
+{
+	g_return_if_fail((path == NULL && style == BACKDROP_NONE) ||
+			 (path != NULL && style != BACKDROP_NONE));
+
+	if (!current_pinboard)
+	{
+		if (!path)
+			return;
+		pinboard_activate("Default");
+		delayed_error(_("No pinboard was in use... "
+			"the 'Default' pinboard has been selected. "
+			"Use 'rox -p=Default' to turn it on in "
+			"future."));
+		g_return_if_fail(current_pinboard != NULL);
+	}
+
+	/* We might have just run the old backdrop program and now
+	 * we're going to set a new one! Seems a bit mean...
+	 */
+
+	abandon_backdrop_app(current_pinboard);
+
+	g_free(current_pinboard->backdrop);
+	current_pinboard->backdrop = g_strdup(path);
+	current_pinboard->backdrop_style = style;
+	reload_backdrop(current_pinboard,
+			current_pinboard->backdrop,
+			current_pinboard->backdrop_style);
+	
+	pinboard_save();
+
+	if (set_backdrop_dialog)
+	{
+		DropBox *box = g_object_get_data(G_OBJECT(set_backdrop_dialog),
+							  "rox-dropbox");
+		g_return_if_fail(box != NULL);
+		drop_box_set_path(box, current_pinboard->backdrop);
+		update_radios(set_backdrop_dialog);
+	}
+}
+
+/* Called on xrandr screen resizes */
+void pinboard_update_size(void)
+{
+	int width, height;
+
+	gtk_window_get_size(GTK_WINDOW(current_pinboard->window),
+			&width, &height);
+	
+	/* Only update the pinboard's size if the screen gets bigger,
+	 * not smaller. Not sure what to do about icons that end up
+	 * offscreen if the screen shrinks, but perhaps a good policy
+	 * is to leave them there until the screen size is increased
+	 * again rather than mess them around. */
+	width = MAX(width, screen_width);
+	height = MAX(height, screen_height);
+	
+	gtk_widget_set_size_request(current_pinboard->window, width, height);
+}
+
 /****************************************************************
  *			INTERNAL FUNCTIONS			*
  ****************************************************************/
@@ -671,7 +740,7 @@ static void backdrop_response(GtkWidget *dialog, gint response, gpointer data)
 
 static void clear_backdrop(GtkWidget *drop_box, gpointer data)
 {
-	set_backdrop(NULL, BACKDROP_NONE);
+	pinboard_set_backdrop(NULL, BACKDROP_NONE);
 }
 
 static void drag_backdrop_dropped(GtkWidget	*drop_box,
@@ -698,7 +767,7 @@ static void drag_backdrop_dropped(GtkWidget	*drop_box,
 		pinboard_set_backdrop_app(path);
 	}
 	else if (S_ISREG(info.st_mode))
-		set_backdrop(path, radios_get_value(radios));
+		pinboard_set_backdrop(path, radios_get_value(radios));
 	else
 		delayed_error(_("Only files (and certain applications) can be "
 				"used to set the background image."));
@@ -1309,7 +1378,7 @@ static const char *pin_from_file(gchar *line)
 
 		line = end + 1;
 
-		while (isspace(*line))
+		while (g_ascii_isspace(*line))
 			line++;
 		if (*line != ',')
 			return _("Missing ',' after icon label");
@@ -1933,13 +2002,17 @@ static void pin_icon_set_tip(PinIcon *pi)
 static void pinboard_show_menu(GdkEventButton *event, PinIcon *pi)
 {
 	int		pos[3];
+	GList		*list;
 
 	pos[0] = event->x_root;
 	pos[1] = event->y_root;
-	pos[2] = 1;
 
 	icon_prepare_menu((Icon *) pi, TRUE);
 
+	list = gtk_container_get_children(GTK_CONTAINER(icon_menu));
+	pos[2] = g_list_length(list) - 6;
+	g_list_free(list);
+
 	gtk_menu_popup(GTK_MENU(icon_menu), NULL, NULL,
 			position_menu,
 			(gpointer) pos, event->button, event->time);
@@ -2028,7 +2101,7 @@ static GdkPixmap *load_backdrop(const gchar *path, BackdropStyle style)
 				"Backdrop removed."),
 				error->message);
 		g_error_free(error);
-		set_backdrop(NULL, BACKDROP_NONE);
+		pinboard_set_backdrop(NULL, BACKDROP_NONE);
 		return NULL;
 	}
 
@@ -2288,52 +2361,6 @@ static void reload_backdrop(Pinboard *pinboard,
 	}
 }
 
-/* Set and save (path, style) as the new backdrop.
- * If style is BACKDROP_PROGRAM, the program is run to get the backdrop.
- * Otherwise, the image is displayed now.
- */
-static void set_backdrop(const gchar *path, BackdropStyle style)
-{
-	g_return_if_fail((path == NULL && style == BACKDROP_NONE) ||
-			 (path != NULL && style != BACKDROP_NONE));
-
-	if (!current_pinboard)
-	{
-		if (!path)
-			return;
-		pinboard_activate("Default");
-		delayed_error(_("No pinboard was in use... "
-			"the 'Default' pinboard has been selected. "
-			"Use 'rox -p=Default' to turn it on in "
-			"future."));
-		g_return_if_fail(current_pinboard != NULL);
-	}
-
-	/* We might have just run the old backdrop program and now
-	 * we're going to set a new one! Seems a bit mean...
-	 */
-
-	abandon_backdrop_app(current_pinboard);
-
-	g_free(current_pinboard->backdrop);
-	current_pinboard->backdrop = g_strdup(path);
-	current_pinboard->backdrop_style = style;
-	reload_backdrop(current_pinboard,
-			current_pinboard->backdrop,
-			current_pinboard->backdrop_style);
-	
-	pinboard_save();
-
-	if (set_backdrop_dialog)
-	{
-		DropBox *box = g_object_get_data(G_OBJECT(set_backdrop_dialog),
-							  "rox-dropbox");
-		g_return_if_fail(box != NULL);
-		drop_box_set_path(box, current_pinboard->backdrop);
-		update_radios(set_backdrop_dialog);
-	}
-}
-
 #define SEARCH_STEP 32
 
 static void search_free(GdkRectangle *rect, GdkRegion *used,
@@ -2375,6 +2402,26 @@ static void find_free_rect(Pinboard *pinboard, GdkRectangle *rect)
 
 	panel_mark_used(used);
 
+	/* Subtract the no-go areas... */
+
+	if (o_top_margin.int_value > 0)
+	{
+		used_rect.x = 0;
+		used_rect.y = 0;
+		used_rect.width = gdk_screen_width();
+		used_rect.height = o_top_margin.int_value;
+		gdk_region_union_with_rect(used, &used_rect);
+	}
+
+	if (o_bottom_margin.int_value > 0)
+	{
+		used_rect.x = 0;
+		used_rect.y = gdk_screen_height() - o_bottom_margin.int_value;
+		used_rect.width = gdk_screen_width();
+		used_rect.height = o_bottom_margin.int_value;
+		gdk_region_union_with_rect(used, &used_rect);
+	}
+
 	/* Subtract the used areas... */
 
 	next = GTK_FIXED(pinboard->fixed)->children;
@@ -2474,7 +2521,7 @@ static void radios_changed(gpointer data)
 	{
 		path = drop_box_get_path(drop_box);
 		if (path)
-			set_backdrop(path, radios_get_value(radios));
+			pinboard_set_backdrop(path, radios_get_value(radios));
 	}
 }
 
