@@ -1,8 +1,8 @@
 /*
- * $Id: remote.c,v 1.19.2.1 2002/07/02 21:38:02 tal197 Exp $
+ * $Id: remote.c,v 1.54 2004/03/02 17:50:46 kerofin Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2002, the ROX-Filer team.
+ * Copyright (C) 2003, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -48,6 +48,9 @@
 #include "panel.h"
 #include "action.h"
 #include "type.h"
+#include "display.h"
+#include "xml.h"
+#include "diritem.h"
 
 static GdkAtom filer_atom;	/* _ROX_FILER_EUID_VERSION_HOST */
 static GdkAtom filer_atom_any;	/* _ROX_FILER_EUID_HOST */
@@ -65,12 +68,15 @@ struct _SOAP_call {
 static GHashTable *rpc_calls = NULL; /* MethodName -> Function */
 
 /* Static prototypes */
-static GdkWindow *get_existing_ipc_window();
+static GdkWindow *get_existing_ipc_window(void);
 static gboolean get_ipc_property(GdkWindow *window, Window *r_xid);
 static void soap_send(GtkWidget *from, GdkAtom prop, GdkWindow *dest);
 static gboolean client_event(GtkWidget *window,
 				 GdkEventClient *event,
 				 gpointer data);
+static void soap_done(GtkWidget *widget,
+		      GdkEventProperty *event,
+		      gpointer data);
 static void soap_register(char *name, SOAP_func func, char *req, char *opt);
 static xmlNodePtr soap_invoke(xmlNode *method);
 
@@ -88,6 +94,11 @@ static xmlNodePtr rpc_Link(GList *args);
 static xmlNodePtr rpc_FileType(GList *args);
 static xmlNodePtr rpc_Mount(GList *args);
 
+static xmlNodePtr rpc_PanelAdd(GList *args);
+static xmlNodePtr rpc_PinboardAdd(GList *args);
+static xmlNodePtr rpc_SetBackdrop(GList *args);
+static xmlNodePtr rpc_SetBackdropApp(GList *args);
+
 /****************************************************************
  *			EXTERNAL INTERFACE			*
  ****************************************************************/
@@ -111,7 +122,8 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 	soap_register("Version", rpc_Version, NULL, NULL);
 
 	soap_register("Run", rpc_Run, "Filename", NULL);
-	soap_register("OpenDir", rpc_OpenDir, "Filename", NULL);
+	soap_register("OpenDir", rpc_OpenDir, "Filename",
+			      "Style,Details,Sort,Class,ID,Hidden,Filter");
 	soap_register("CloseDir", rpc_CloseDir, "Filename", NULL);
 	soap_register("Examine", rpc_Examine, "Filename", NULL);
 	soap_register("Show", rpc_Show, "Directory,Leafname", NULL);
@@ -126,6 +138,11 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 	soap_register("Link", rpc_Link, "From,To", "Leafname");
 	soap_register("Mount", rpc_Mount, "MountPoints", "OpenDir,Quiet");
 
+	soap_register("SetBackdrop", rpc_SetBackdrop, "Filename,Style", NULL);
+	soap_register("SetBackdropApp", rpc_SetBackdropApp, "App", NULL);
+	soap_register("PinboardAdd", rpc_PinboardAdd, "Path,X,Y", "Label");
+	soap_register("PanelAdd", rpc_PanelAdd, "Side,Path", "Label,After");
+
 	/* Look for a property on the root window giving the IPC window
 	 * of an already-running copy of this version of the filer, running
 	 * on the same machine and with the same euid.
@@ -146,15 +163,26 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 	ipc_window = gtk_invisible_new();
 	gtk_widget_realize(ipc_window);
 
+	XGrabServer(GDK_DISPLAY());
+
 	existing_ipc_window = new_copy ? NULL : get_existing_ipc_window();
 	if (existing_ipc_window)
 	{
 		xmlChar *mem;
 		int	size;
 
+		XUngrabServer(GDK_DISPLAY());
+
 		xmlDocDumpMemory(rpc, &mem, &size);
 		g_return_val_if_fail(size > 0, FALSE);
 		
+		/* Since Gtk might have selected this already, we'd
+		 * better do it BEFORE changing the property.
+		 */
+		gtk_widget_add_events(ipc_window, GDK_PROPERTY_CHANGE_MASK);
+		g_signal_connect(ipc_window, "property-notify-event",
+				G_CALLBACK(soap_done), GINT_TO_POINTER(xsoap));
+
 		gdk_property_change(ipc_window->window, xsoap,
 				gdk_x11_xatom_to_atom(XA_STRING), 8,
 				GDK_PROP_MODE_REPLACE, mem, size);
@@ -177,15 +205,17 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 			(void *) &xwindow, 1);
 
 	/* Get notified when we get a message */
-	gtk_signal_connect(GTK_OBJECT(ipc_window), "client-event",
-			GTK_SIGNAL_FUNC(client_event), NULL);
+	g_signal_connect(ipc_window, "client-event",
+			G_CALLBACK(client_event), NULL);
 
 	/* Make the root window contain a pointer to the IPC window */
-	gdk_property_change(GDK_ROOT_PARENT(), filer_atom,
+	gdk_property_change(gdk_get_default_root_window(), filer_atom,
 			gdk_x11_xatom_to_atom(XA_WINDOW), 32,
 			GDK_PROP_MODE_REPLACE,
 			(void *) &xwindow, 1);
 
+	XUngrabServer(GDK_DISPLAY());
+
 	/* Also have a property without the version number, for programs
 	 * that are happy to talk to any version of the filer.
 	 */
@@ -199,7 +229,7 @@ gboolean remote_init(xmlDocPtr rpc, gboolean new_copy)
 			GDK_PROP_MODE_REPLACE,
 			(void *) &xwindow, 1);
 	/* ... and on the root */
-	gdk_property_change(GDK_ROOT_PARENT(), filer_atom_any,
+	gdk_property_change(gdk_get_default_root_window(), filer_atom_any,
 			gdk_x11_xatom_to_atom(XA_WINDOW), 32,
 			GDK_PROP_MODE_REPLACE,
 			(void *) &xwindow, 1);
@@ -301,7 +331,7 @@ static GdkWindow *get_existing_ipc_window(void)
 	Window		xid, xid_confirm;
 	GdkWindow	*window;
 
-	if (!get_ipc_property(GDK_ROOT_PARENT(), &xid))
+	if (!get_ipc_property(gdk_get_default_root_window(), &xid))
 		return NULL;
 
 	if (gdk_window_lookup(xid))
@@ -396,7 +426,12 @@ static gboolean client_event(GtkWidget *window,
 		return FALSE;
 
 	src_window = gdk_window_foreign_new(event->data.l[0]);
-	g_return_val_if_fail(src_window != NULL, FALSE);
+	if (!src_window)
+	{
+		g_warning("SOAP message sender window was destroyed before I \n"
+			  "could read it.");
+		return FALSE;
+	}
 	prop = gdk_x11_xatom_to_atom(event->data.l[1]);
 
 	data = read_property(src_window, prop, &length);
@@ -476,6 +511,26 @@ static char *string_value(xmlNode *arg)
 	return retval ? retval : g_strdup("");
 }
 
+/* Returns the text of this arg as an int, or the default value if not
+ * supplied or not an int.
+ */
+static int int_value(xmlNode *arg, int def)
+{
+	char *str, *end;
+	int i;
+
+	if (!arg)
+		return def;
+	
+	str = xmlNodeGetContent(arg);
+	if (!str || !str[0])
+		return def;
+
+	i = (int) strtol(str, &end, 0);
+
+	return (end > str) ? i : def;
+}
+
 /* Return a list of strings, one for each child node of arg.
  * g_list_free the list, and g_free each string.
  */
@@ -508,18 +563,123 @@ static xmlNodePtr rpc_Version(GList *args)
 
 	reply = xmlNewNode(NULL, "rox:VersionResponse");
 	xmlNewNs(reply, SOAP_RPC_NS, "soap");
-	xmlNewChild(reply, NULL, "soap:result", VERSION);
+	xmlNewTextChild(reply, NULL, "soap:result", VERSION);
 
 	return reply;
 }
 
+/* Args: Path, [Style, Details, Sort, Class, Window] */
 static xmlNodePtr rpc_OpenDir(GList *args)
 {
 	char	   *path;
+	char       *style, *details, *sort, *class, *window;
+	int         hidden=FALSE;
+	char       *filter_string=NULL;
+	FilerWindow *fwin = NULL;
 
 	path = string_value(ARG(0));
-	filer_opendir(path, NULL);
+	class = string_value(ARG(4));
+	window = string_value(ARG(5));
+	
+	if (window)
+		fwin = filer_get_by_id(window);
+
+	if (!fwin)
+	{
+		fwin = filer_opendir(path, NULL, class);
+		if (window)
+			filer_set_id(fwin, window); 
+	}
+	else
+		filer_change_to(fwin, path, NULL);
+
 	g_free(path);
+	g_free(class);
+	g_free(window);
+	if (!fwin)
+		return NULL;
+
+	style = string_value(ARG(1));
+	details = string_value(ARG(2));
+	sort = string_value(ARG(3));
+	hidden = bool_value(ARG(6));
+	filter_string = string_value(ARG(7));
+
+	if (style)
+	{
+		DisplayStyle ds;
+
+		ds = !g_strcasecmp(style, "Large") ? LARGE_ICONS :
+		     !g_strcasecmp(style, "Small") ? SMALL_ICONS :
+		     !g_strcasecmp(style, "Huge")  ? HUGE_ICONS :
+		     !g_strcasecmp(style, "Automatic")  ? AUTO_SIZE_ICONS :
+		     				     UNKNOWN_STYLE;
+		if (ds == UNKNOWN_STYLE)
+			delayed_error(_("Unknown style '%s'"), style);
+		else
+			display_set_layout(fwin, ds, fwin->details_type, TRUE);
+
+		g_free(style);
+	}
+
+	if (details)
+	{
+		DetailsType dt;
+		ViewType view_type;
+		
+		dt = !g_strcasecmp(details, "None") ? DETAILS_NONE :
+		     !g_strcasecmp(details, "ListView") ? DETAILS_NONE :
+		     !g_strcasecmp(details, "Size") ? DETAILS_SIZE :
+		     !g_strcasecmp(details, "Type") ? DETAILS_TYPE :
+		     !g_strcasecmp(details, "Times") ? DETAILS_TIMES :
+		     !g_strcasecmp(details, "Permissions")
+		     				? DETAILS_PERMISSIONS :
+						  DETAILS_UNKNOWN;
+
+		if (dt == DETAILS_UNKNOWN)
+			delayed_error(_("Unknown details type '%s'"), details);
+		else
+			display_set_layout(fwin, fwin->display_style_wanted,
+					   dt, TRUE);
+
+		if (g_strcasecmp(details, "ListView") == 0)
+			view_type = VIEW_TYPE_DETAILS;
+		else
+			view_type = VIEW_TYPE_COLLECTION;
+
+		if (view_type != fwin->view_type)
+			filer_set_view_type(fwin, view_type);
+		
+		g_free(details);
+	}
+
+	if (sort)
+	{
+		SortType type;
+
+		type = !g_strcasecmp(sort, "Name") ? SORT_NAME :
+		       !g_strcasecmp(sort, "Type") ? SORT_TYPE :
+		       !g_strcasecmp(sort, "Date") ? SORT_DATE :
+ 		       !g_strcasecmp(sort, "Size") ? SORT_SIZE :
+ 		       !g_strcasecmp(sort, "Owner") ? SORT_OWNER :
+ 		       !g_strcasecmp(sort, "Group") ? SORT_GROUP :
+						       -1;
+		if (type == -1)
+			delayed_error(_("Unknown sorting type '%s'"), sort);
+		else
+			display_set_sort_type(fwin, type, GTK_SORT_ASCENDING);
+
+		g_free(sort);
+	}
+	if (hidden!=-1)
+	{
+		display_set_hidden(fwin, hidden);
+	}
+
+	if (filter_string)
+		display_set_filter(fwin, FILER_SHOW_GLOB, filter_string);
+	else
+		display_set_filter(fwin, FILER_SHOW_ALL, NULL);
 
 	return NULL;
 }
@@ -565,7 +725,7 @@ static xmlNodePtr rpc_Show(GList *args)
 	leaf = string_value(ARG(1));
 
 	/* XXX: Seems silly to join them only to split again later... */
-	open_to_show(make_path(dir, leaf)->str);
+	open_to_show(make_path(dir, leaf));
 
 	g_free(dir);
 	g_free(leaf);
@@ -584,31 +744,128 @@ static xmlNodePtr rpc_Pinboard(GList *args)
 	return NULL;
 }
 
+/* args = Filename, Style */
+static xmlNodePtr rpc_SetBackdrop(GList *args)
+{
+	char *file;
+	char *style;
+	BackdropStyle s;
+
+	file = string_value(ARG(0));
+	style = string_value(ARG(1));
+
+	s = !g_strcasecmp(style, "Tile") ? BACKDROP_TILE :
+	    !g_strcasecmp(style, "Scale") ? BACKDROP_SCALE :
+	    !g_strcasecmp(style, "Stretch") ? BACKDROP_STRETCH :
+	    !g_strcasecmp(style, "Centre") ? BACKDROP_CENTRE :
+					     BACKDROP_NONE;
+
+	if (s == BACKDROP_NONE)
+		g_warning("Invalid style '%s' for backdrop", style);
+	else
+		pinboard_set_backdrop(file, s);
+
+	g_free(file);
+	g_free(style);
+
+	return NULL;
+}
+
+/* args = App */
+static xmlNodePtr rpc_SetBackdropApp(GList *args)
+{
+	char *app;
+
+	app = string_value(ARG(0));
+
+	pinboard_set_backdrop_app(app);
+
+	g_free(app);
+
+	return NULL;
+}
+
+/* args = Path, X, Y, [Label] */
+static xmlNodePtr rpc_PinboardAdd(GList *args)
+{
+	char *path = NULL;
+	gchar *name;
+	int x, y;
+
+	path = string_value(ARG(0));
+	x = int_value(ARG(1), 0);
+	y = int_value(ARG(2), 0);
+	name = string_value(ARG(3));
+	
+	pinboard_pin(path, name, x, y, NULL);
+
+	g_free(path);
+
+	return NULL;
+}
+
+/* Returns PANEL_NUMBER_OF_SIDES if name is invalid */
+static PanelSide panel_name_to_side(gchar *side)
+{
+	if (strcmp(side, "Top") == 0)
+		return PANEL_TOP;
+	else if (strcmp(side, "Bottom") == 0)
+		return PANEL_BOTTOM;
+	else if (strcmp(side, "Left") == 0)
+		return PANEL_LEFT;
+	else if (strcmp(side, "Right") == 0)
+		return PANEL_RIGHT;
+	else
+		g_warning("Unknown panel side '%s'", side);
+	return PANEL_NUMBER_OF_SIDES;
+}
+
 /* args = Side, [Name] */
 static xmlNodePtr rpc_Panel(GList *args)
 {
-	char *name, *side;
+	PanelSide side;
+	char *name, *side_name;
+
+	side_name = string_value(ARG(0));
+	side = panel_name_to_side(side_name);
+	g_free(side_name);
+	g_return_val_if_fail(side != PANEL_NUMBER_OF_SIDES, NULL);
 
-	side = string_value(ARG(0));
 	name = string_value(ARG(1));
 
-	if (strcmp(side, "Top") == 0)
-		panel_new(name, PANEL_TOP);
-	else if (strcmp(side, "Bottom") == 0)
-		panel_new(name, PANEL_BOTTOM);
-	else if (strcmp(side, "Left") == 0)
-		panel_new(name, PANEL_LEFT);
-	else if (strcmp(side, "Right") == 0)
-		panel_new(name, PANEL_RIGHT);
-	else
-		g_warning("Unknown panel side '%s'", side);
+	panel_new(name, side);
 
-	g_free(side);
 	g_free(name);
 
 	return NULL;
 }
 
+/* args = Side, Path, [Label, After] */
+static xmlNodePtr rpc_PanelAdd(GList *args)
+{
+	PanelSide side;
+	char *path, *side_name, *label;
+	gboolean after = FALSE;
+	int tmp;
+
+	side_name = string_value(ARG(0));
+	side = panel_name_to_side(side_name);
+	g_free(side_name);
+	g_return_val_if_fail(side != PANEL_NUMBER_OF_SIDES, NULL);
+
+	path = string_value(ARG(1));
+	label = string_value(ARG(2));
+
+	tmp = bool_value(ARG(3));
+	after = (tmp == -1) ? FALSE : tmp;
+
+	panel_add(side, path, label, after);
+
+	g_free(path);
+
+	return NULL;
+}
+
 static xmlNodePtr rpc_Copy(GList *args)
 {
 	GList *from;
@@ -623,9 +880,10 @@ static xmlNodePtr rpc_Copy(GList *args)
 
 	if (from)
 		action_copy(from, to, leaf, quiet);
+	else
+		g_warning("No files in SOAP request list");
 
-	g_list_foreach(from, (GFunc) g_free, NULL);
-	g_list_free(from);
+	destroy_glist(&from);
 	g_free(to);
 	g_free(leaf);
 
@@ -646,9 +904,10 @@ static xmlNodePtr rpc_Move(GList *args)
 
 	if (from)
 		action_move(from, to, leaf, quiet);
+	else
+		g_warning("No files in SOAP request list");
 
-	g_list_foreach(from, (GFunc) g_free, NULL);
-	g_list_free(from);
+	destroy_glist(&from);
 	g_free(to);
 	g_free(leaf);
 
@@ -667,9 +926,10 @@ static xmlNodePtr rpc_Link(GList *args)
 
 	if (from)
 		action_link(from, to, leaf);
+	else
+		g_warning("No files in SOAP request list");
 
-	g_list_foreach(from, (GFunc) g_free, NULL);
-	g_list_free(from);
+	destroy_glist(&from);
 	g_free(to);
 	g_free(leaf);
 
@@ -690,7 +950,7 @@ static xmlNodePtr rpc_FileType(GList *args)
 	tname = g_strconcat(type->media_type, "/", type->subtype, NULL);
 
 	xmlNewNs(reply, SOAP_RPC_NS, "soap");
-	xmlNewChild(reply, NULL, "soap:result", tname);
+	xmlNewTextChild(reply, NULL, "soap:result", tname);
 	g_free(tname);
 
 	return reply;
@@ -711,19 +971,29 @@ static xmlNodePtr rpc_Mount(GList *args)
 	if (paths)
 		action_mount(paths, open_dir, quiet);
 
-	g_list_foreach(paths, (GFunc) g_free, NULL);
-	g_list_free(paths);
+	destroy_glist(&paths);
 
 	return NULL;
 }
 
+/* The first time the property changes, do nothing (it's us setting the
+ * property). The second time, get the result.
+ * Don't call this function three times!
+ */
 static void soap_done(GtkWidget *widget, GdkEventProperty *event, gpointer data)
 {
+	static int times_called = 0;
 	GdkAtom	prop = (GdkAtom) data;
 
 	if (prop != event->atom)
 		return;
 
+	times_called++;
+	g_return_if_fail(times_called < 3);
+
+	if (times_called == 1)
+		return;
+
 	/* If we got a reply, display it here */
 	if (event->state == GDK_PROPERTY_NEW_VALUE)
 	{
@@ -738,7 +1008,7 @@ static void soap_done(GtkWidget *widget, GdkEventProperty *event, gpointer data)
 	gtk_main_quit();
 }
 
-gboolean too_slow(gpointer data)
+static gboolean too_slow(gpointer data)
 {
 	g_warning("Existing ROX-Filer process is not responding! Try with -n");
 	gtk_main_quit();
@@ -756,11 +1026,6 @@ static void soap_send(GtkWidget *from, GdkAtom prop, GdkWindow *dest)
 	event.data_format = 32;
 	event.message_type = xsoap;
 	
-	gtk_widget_add_events(from, GDK_PROPERTY_CHANGE_MASK);
-	gtk_signal_connect(GTK_OBJECT(from), "property-notify-event",
-				GTK_SIGNAL_FUNC(soap_done),
-				GINT_TO_POINTER(prop));
-
 	gdk_event_send_client_message((GdkEvent *) &event,
 				      GDK_WINDOW_XWINDOW(dest));
 
@@ -779,6 +1044,8 @@ static xmlNodePtr soap_invoke(xmlNode *method)
 	SOAP_call *call;
 	gchar **arg;
 	xmlNodePtr retval = NULL;
+	GHashTable *name_to_node;
+	xmlNode	*node;
 
 	call = g_hash_table_lookup(rpc_calls, method->name);
 	if (!call)
@@ -791,20 +1058,30 @@ static xmlNodePtr soap_invoke(xmlNode *method)
 		reply = xmlNewNode(NULL, "env:Fault");
 		xmlNewNs(reply, SOAP_RPC_NS, "rpc");
 		xmlNewNs(reply, SOAP_ENV_NS, "env");
-		xmlNewChild(reply, NULL, "faultcode",
+		xmlNewTextChild(reply, NULL, "faultcode",
 						"rpc:ProcedureNotPresent");
-		xmlNewChild(reply, NULL, "faultstring", err);
+		xmlNewTextChild(reply, NULL, "faultstring", err);
 		g_free(err);
 		return reply;
 	}
 
+	name_to_node = g_hash_table_new(g_str_hash, g_str_equal);
+	for (node = method->xmlChildrenNode; node; node = node->next)
+	{
+		if (node->type != XML_ELEMENT_NODE)
+			continue;
+
+		if (node->ns == NULL || strcmp(node->ns->href, ROX_NS) != 0)
+			continue;
+
+		g_hash_table_insert(name_to_node, (gchar *) node->name, node);
+	}
+
 	if (call->required_args)
 	{
 		for (arg = call->required_args; *arg; arg++)
 		{
-			xmlNode *node;
-
-			node = get_subnode(method, ROX_NS, *arg);
+			node = g_hash_table_lookup(name_to_node, *arg);
 			if (!node)
 			{
 				g_warning("Missing required argument '%s' "
@@ -820,17 +1097,14 @@ static xmlNodePtr soap_invoke(xmlNode *method)
 	if (call->optional_args)
 	{
 		for (arg = call->optional_args; *arg; arg++)
-		{
-			xmlNode *node;
-
-			node = get_subnode(method, ROX_NS, *arg);
-
-			args = g_list_append(args, node);
-		}
+			args = g_list_append(args,
+				g_hash_table_lookup(name_to_node, *arg));
 	}
 
 	retval = call->func(args);
+
 out:
+	g_hash_table_destroy(name_to_node);
 	g_list_free(args);
 
 	return retval;
