@@ -1,8 +1,8 @@
 /*
- * $Id: sc.c,v 1.5 2002/01/13 17:47:37 tal197 Exp $
+ * $Id: sc.c,v 1.9 2003/01/19 19:29:04 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2002, the ROX-Filer team.
+ * Copyright (C) 2003, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -31,7 +31,7 @@
 
 #include "sc.h"
 
-static SmPropValue *make_list_of_array_value(gchar *vals[], gint nvals)
+static SmPropValue *make_list_of_array_value(const gchar *vals[], gint nvals)
 {
 	SmPropValue *values = g_new(SmPropValue, nvals);
 	gint i;
@@ -44,7 +44,7 @@ static SmPropValue *make_list_of_array_value(gchar *vals[], gint nvals)
 	return values;
 }
 
-static SmPropValue *make_array_value(gchar *val)
+static SmPropValue *make_array_value(const gchar *val)
 {
 	SmPropValue *value = g_new(SmPropValue, 1);
 	
@@ -64,22 +64,22 @@ static SmPropValue *make_card_value(gchar val)
 	return value;
 }
 
-static SmProperty *find_property(SmClient *client, gchar *name)
+static SmProperty *find_property(SmClient *client, const gchar *name)
 {
 	GSList *list = client->props;
 	SmProperty *prop;
 
-	while(list)
+	for (; list; list = list->next)
 	{
 		prop = (SmProperty *)list->data;
-		if(strcmp(prop->prop.name, name) == 0)
+		if (strcmp(prop->prop.name, name) == 0)
 			return prop;
-		list = g_slist_next(list);		
 	}
 	return NULL;		
 }
 
-static SmProperty *new_property(SmClient *client, gchar *name, gchar *type)
+static SmProperty *new_property(SmClient *client,
+				const gchar *name, const gchar *type)
 {
 	SmProperty *prop;
 
@@ -96,7 +96,7 @@ static SmProperty *new_property(SmClient *client, gchar *name, gchar *type)
 	return prop;
 }
 
-gint close_connection(gpointer data)
+static gint close_connection(gpointer data)
 {
 	SmClient *client = (SmClient *)data;
 	
@@ -108,19 +108,21 @@ gint close_connection(gpointer data)
 /* Called when there's data to be read on the ICE file descriptor.
    Unpacks the message and triggers the correct callback... I think */
 
-void poll_ice_messages(gpointer data, gint source, GdkInputCondition condition)
+static void poll_ice_messages(gpointer data, gint source,
+				GdkInputCondition condition)
 {
 	SmClient *client = (SmClient *)data;
 	Bool ret;
 
-	if(IceProcessMessages(client->ice_conn, NULL, &ret) == IceProcessMessagesIOError)
+	if (IceProcessMessages(client->ice_conn, NULL, &ret) ==
+			IceProcessMessagesIOError)
 		SmcCloseConnection(client->conn, 0, NULL);
 	return;
 }
 
 /* Called whenever an ICE connection is opened or closed */
 
-void ice_watch_fn(IceConn conn, IcePointer client_data,
+static void ice_watch_fn(IceConn conn, IcePointer client_data,
                 			Bool opening, IcePointer *watch_data)
 {
 	SmClient *client = (SmClient *)client_data;
@@ -149,7 +151,7 @@ void ice_watch_fn(IceConn conn, IcePointer client_data,
 
 /* Callbacks for different SM messages */
 
-void sc_save_yourself(SmcConn conn, SmPointer client_data, int save_type,
+static void sc_save_yourself(SmcConn conn, SmPointer client_data, int save_type,
                       		Bool shutdown, int interact_style, Bool fast)
 {
 	SmClient *client = (SmClient *)client_data;
@@ -164,7 +166,7 @@ void sc_save_yourself(SmcConn conn, SmPointer client_data, int save_type,
 	SmcSaveYourselfDone(client->conn, success);
 }
 
-void sc_shutdown_cancelled(SmcConn conn, SmPointer client_data)
+static void sc_shutdown_cancelled(SmcConn conn, SmPointer client_data)
 {
 	SmClient *client = (SmClient *)client_data;
 #ifdef DEBUG	
@@ -174,7 +176,7 @@ void sc_shutdown_cancelled(SmcConn conn, SmPointer client_data)
 		client->shutdown_cancelled_fn(client);
 }
 
-void sc_save_complete(SmcConn conn, SmPointer client_data)
+static void sc_save_complete(SmcConn conn, SmPointer client_data)
 {
 	SmClient *client = (SmClient *)client_data;
 #ifdef DEBUG	
@@ -184,7 +186,7 @@ void sc_save_complete(SmcConn conn, SmPointer client_data)
 		client->save_complete_fn(client);
 }
 
-void sc_die(SmcConn conn, SmPointer client_data)
+static void sc_die(SmcConn conn, SmPointer client_data)
 {
 	SmClient *client = (SmClient *)client_data;
 #ifdef DEBUG	
@@ -201,7 +203,7 @@ gboolean sc_session_up()
 	return TRUE;
 }
 
-SmClient *sc_new(gchar *client_id)
+SmClient *sc_new(const gchar *client_id)
 {
 	SmClient *client;
 	
@@ -226,7 +228,7 @@ void sc_destroy(SmClient *client)
 #ifdef DEBUG
 	g_printerr("destroying client\n");
 #endif	
-	while(list)
+	for (; list; list = list->next)
 	{
 		prop = (SmProperty *)list->data;
 		g_free(prop->prop.vals->value);
@@ -234,7 +236,6 @@ void sc_destroy(SmClient *client)
 		g_free(prop->prop.name);
 		g_free(prop->prop.type);
 		g_free(prop);
-		list = g_slist_next(list);
 	}
 	g_slist_free(client->props);
 	g_free(client->id);
@@ -244,7 +245,9 @@ void sc_destroy(SmClient *client)
 /* Set a property with a SmLISTofARRAY8 value as in SmRestartCommand,
    SmCloneCommand and SmDiscardCommand */ 
 
-void sc_set_list_of_array_prop(SmClient *client, gchar *name, gchar *vals[], gint nvals)
+void sc_set_list_of_array_prop(SmClient *client,
+			const gchar *name, 
+			const char *vals[], gint nvals)
 {
 	SmProperty *prop = find_property(client, name);
 	SmPropValue *value = make_list_of_array_value(vals, nvals);
@@ -263,7 +266,7 @@ void sc_set_list_of_array_prop(SmClient *client, gchar *name, gchar *vals[], gin
 /* Set a prop with a SmARRAY8 value, SmProgram, SmUserID and
    SmDiscardCommand (if using XSM) are suitable victims. */
 
-void sc_set_array_prop(SmClient *client, gchar *name, gchar *vals)
+void sc_set_array_prop(SmClient *client, const gchar *name, const gchar *vals)
 {
 	SmProperty *prop = find_property(client, name);
 	SmPropValue *value = make_array_value(vals);
@@ -281,7 +284,7 @@ void sc_set_array_prop(SmClient *client, gchar *name, gchar *vals)
 
 /* This one is for the SmRestarStyleHint */
 
-void sc_set_card_prop(SmClient *client, gchar *name, gchar val)
+void sc_set_card_prop(SmClient *client, const gchar *name, const gchar val)
 {
 	SmProperty *prop = find_property(client, name);
 	SmPropValue *value = make_card_value(val);
@@ -307,7 +310,8 @@ void sc_set_card_prop(SmClient *client, gchar *name, gchar val)
   The values are those stored in the SmClient struct,
   They're not fetched from the session manager */
 
-void sc_get_prop_value(SmClient *client, gchar *name, SmPropValue **val_ret, gint *nvals_ret)
+void sc_get_prop_value(SmClient *client, const gchar *name,
+			SmPropValue **val_ret, gint *nvals_ret)
 {
 	SmProperty *prop = find_property(client, name);
 	
@@ -334,7 +338,7 @@ void sc_register_properties(SmClient *client)
 	gint i;
 #endif
 	
-	while(list)
+	for (; list; list = list->next)
 	{
 		prop = (SmProperty *)list->data;
 		if(prop->set == TRUE)
@@ -342,7 +346,6 @@ void sc_register_properties(SmClient *client)
 			g_ptr_array_add(set_props, &prop->prop);
 			prop->set = FALSE;
 		}
-		list = g_slist_next(list);
 	}
 #ifdef DEBUG
 	g_printerr("Registering props:\n");
