@@ -1,8 +1,8 @@
 /*
- * $Id: support.c,v 1.72.2.1 2002/07/02 21:38:02 tal197 Exp $
+ * $Id: support.c,v 1.111 2004/04/27 10:34:30 tal197 Exp $
  *
  * ROX-Filer, filer for the ROX desktop project
- * Copyright (C) 2002, the ROX-Filer team.
+ * Copyright (C) 2003, the ROX-Filer team.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -37,6 +37,7 @@
 #include <time.h>
 #include <unistd.h>
 #include <libxml/parser.h>
+#include <math.h>
 
 #include "global.h"
 
@@ -44,78 +45,32 @@
 #include "main.h"
 #include "options.h"
 #include "support.h"
-#include "my_vfs.h"
 #include "fscache.h"
-
-#ifndef GTK2
-# include "gconvert.h"
-#endif
+#include "main.h"
+#include "xml.h"
 
 static GHashTable *uid_hash = NULL;	/* UID -> User name */
 static GHashTable *gid_hash = NULL;	/* GID -> Group name */
 
 /* Static prototypes */
-#if defined(GTK2) || defined(THUMBS_USE_LIBPNG)
 static void MD5Transform(guint32 buf[4], guint32 const in[16]);
-#endif
-
-/* Static prototypes */
-static XMLwrapper *xml_load(char *pathname, gpointer data);
-static void xml_ref(XMLwrapper *dir, gpointer data);
-static void xml_unref(XMLwrapper *dir, gpointer data);
-static int xml_getref(XMLwrapper *dir, gpointer data);
 
 /****************************************************************
  *			EXTERNAL INTERFACE			*
  ****************************************************************/
 
-XMLwrapper *xml_cache_load(gchar *pathname)
+/* g_object_unref() the result! */
+XMLwrapper *xml_cache_load(const gchar *pathname)
 {
 	static GFSCache *xml_cache = NULL;
 
 	if (!xml_cache)
-		xml_cache = g_fscache_new((GFSLoadFunc) xml_load,
-					  (GFSRefFunc) xml_ref,
-					  (GFSRefFunc) xml_unref,
-					  (GFSGetRefFunc) xml_getref,
-					  NULL, NULL);
+		xml_cache = g_fscache_new((GFSLoadFunc) xml_new, NULL, NULL);
 	return g_fscache_lookup(xml_cache, pathname);
 }
 
-void xml_cache_unref(XMLwrapper *wrapper)
-{
-	xml_unref(wrapper, NULL);
-}
-
-/* Return the (first) child of this node with the given name.
- * NULL if not found.
- */
-xmlNode *get_subnode(xmlNode *node, const char *namespaceURI, const char *name)
-{
-	for (node = node->xmlChildrenNode; node; node = node->next)
-	{
-		if (node->type != XML_ELEMENT_NODE)
-			continue;
-
-		if (strcmp(node->name, name))
-			continue;
-
-		if (node->ns == NULL || namespaceURI == NULL)
-		{
-			if (node->ns == NULL && namespaceURI == NULL)
-				return node;
-			continue;
-		}
-		
-		if (strcmp(node->ns->href, namespaceURI) == 0)
-			return node;
-	}
-
-	return NULL;
-}
-
-/* Save doc as XML as filename, 0 on success or -1 on failure */
-int save_xml_file(xmlDocPtr doc, gchar *filename)
+/* Save doc as XML as filename, 0 on success or 1 on failure */
+int save_xml_file(xmlDocPtr doc, const gchar *filename)
 {
 #if LIBXML_VERSION > 20400
 	if (xmlSaveFormatFileEnc(filename, doc, NULL, 1) < 0)
@@ -159,7 +114,7 @@ xmlDocPtr soap_new(xmlNodePtr *ret_body)
 }
 
 /* Like g_strdup, but does realpath() too (if possible) */
-char *pathdup(char *path)
+char *pathdup(const char *path)
 {
 	char real[MAXPATHLEN];
 
@@ -172,37 +127,41 @@ char *pathdup(char *path)
 }
 
 /* Join the path to the leaf (adding a / between them) and
- * return a pointer to a buffer with the result. Buffer is valid until
- * the next call to make_path.
+ * return a pointer to a static buffer with the result. Buffer is valid
+ * until the next call to make_path.
+ * The return value may be used as 'dir' for the next call.
  */
-GString *make_path(char *dir, char *leaf)
+const guchar *make_path(const char *dir, const char *leaf)
 {
 	static GString *buffer = NULL;
 
 	if (!buffer)
 		buffer = g_string_new(NULL);
 
-	g_return_val_if_fail(dir != NULL, buffer);
-	g_return_val_if_fail(leaf != NULL, buffer);
+	g_return_val_if_fail(dir != NULL, buffer->str);
+	g_return_val_if_fail(leaf != NULL, buffer->str);
 
-	g_string_sprintf(buffer, "%s%s%s",
-			dir,
-			dir[0] == '/' && dir[1] == '\0' ? "" : "/",
-			leaf);
+	if (buffer->str != dir)
+		g_string_assign(buffer, dir);
 
-	return buffer;
+	if (dir[0] != '/' || dir[1] != '\0')
+		g_string_append_c(buffer, '/');	/* For anything except "/" */
+
+	g_string_append(buffer, leaf);
+
+	return buffer->str;
 }
 
 /* Return our complete host name for DND */
-char *our_host_name_for_dnd(void)
+const char *our_host_name_for_dnd(void)
 {
-	if (option_get_int("dnd_no_hostnames"))
+	if (o_dnd_no_hostnames.int_value)
 		return "";
 	return our_host_name();
 }
 
 /* Return our complete host name, unconditionally */
-char *our_host_name(void)
+const char *our_host_name(void)
 {
 	static char *name = NULL;
 
@@ -229,72 +188,15 @@ char *our_host_name(void)
 	return name;
 }
 
-/* Create a child process. cd to dir first (if dir is non-NULL).
- * If from_stderr is set, create a pipe for stderr and return the readable
- * side here.
- * Returns the PID of the child, or 0 on failure (from_stderr is still valid).
- */
-pid_t spawn_full(char **argv, char *dir, int *from_stderr)
-{
-	int	child;
-	int	fd[2];
-
-	if (from_stderr)
-	{
-		if (pipe(fd) == 0)
-			*from_stderr = fd[0];
-		else
-		{
-			*from_stderr = -1;
-			from_stderr = NULL;
-		}
-	}
-
-	child = fork();
-
-	if (child == -1)
-		return 0;	/* Failure */
-	else if (child == 0)
-	{
-		/* We are the child process */
-		if (from_stderr)
-		{
-			close(fd[0]);
-			if (fd[1] != STDERR_FILENO)
-			{
-				dup2(fd[1], STDERR_FILENO);
-				close(fd[1]);
-				close_on_exec(STDERR_FILENO, FALSE);
-			}
-		}
-
-		if (dir)
-			if (chdir(dir))
-				fprintf(stderr, "chdir() failed: %s\n",
-						g_strerror(errno));
-		execvp(argv[0], argv);
-		fprintf(stderr, "execvp(%s, ...) failed: %s\n",
-				argv[0],
-				g_strerror(errno));
-		_exit(0);
-	}
-
-	if (from_stderr)
-		close(fd[1]);
-
-	/* We are the parent */
-	return child;
-}
-
 void debug_free_string(void *data)
 {
 	g_print("Freeing string '%s'\n", (char *) data);
 	g_free(data);
 }
 
-char *user_name(uid_t uid)
+const char *user_name(uid_t uid)
 {
-	char	*retval;
+	const char *retval;
 	
 	if (!uid_hash)
 		uid_hash = g_hash_table_new(NULL, NULL);
@@ -308,15 +210,16 @@ char *user_name(uid_t uid)
 		passwd = getpwuid(uid);
 		retval = passwd ? g_strdup(passwd->pw_name)
 			       : g_strdup_printf("[%d]", (int) uid);
-		g_hash_table_insert(uid_hash, GINT_TO_POINTER(uid), retval);
+		g_hash_table_insert(uid_hash, GINT_TO_POINTER(uid),
+				(gchar *) retval);
 	}
 
 	return retval;
 }
 
-char *group_name(gid_t gid)
+const char *group_name(gid_t gid)
 {
-	char	*retval;
+	const char *retval;
 	
 	if (!gid_hash)
 		gid_hash = g_hash_table_new(NULL, NULL);
@@ -330,19 +233,20 @@ char *group_name(gid_t gid)
 		group = getgrgid(gid);
 		retval = group ? g_strdup(group->gr_name)
 			       : g_strdup_printf("[%d]", (int) gid);
-		g_hash_table_insert(gid_hash, GINT_TO_POINTER(gid), retval);
+		g_hash_table_insert(gid_hash, GINT_TO_POINTER(gid),
+				(gchar *) retval);
 	}
 
 	return retval;
 }
 
-/* Return a string in the form '23Mb' in a static buffer valid until
+/* Return a string in the form '23 M' in a static buffer valid until
  * the next call.
  */
-char *format_size(off_t size)
+const char *format_size(off_t size)
 {
 	static	char *buffer = NULL;
-	char	*units;
+	const char	*units;
 	
 	if (size >= PRETTY_SIZE_LIMIT)
 	{
@@ -356,30 +260,27 @@ char *format_size(off_t size)
 			{
 				size += 1023;
 				size >>= 10;
-				units = "Gb";
+				units = "G";
 			}
 			else
-				units = "Mb";
+				units = "M";
 		}
 		else
 			units = "K";
 	}
-	else if (size == 1)
-		units = _("byte");
 	else
-		units = _("bytes");
+		units = _("B");
 
-	if (buffer)
-		g_free(buffer);
+	g_free(buffer);
 	buffer = g_strdup_printf("%" SIZE_FMT " %s", size, units);
 
 	return buffer;
 }
 
-/* Return a string in the form '23Mb' in a static buffer valid until
+/* Return a string in the form '23M' in a static buffer valid until
  * the next call. Aligned to the right (5 chars).
  */
-char *format_size_aligned(off_t size)
+const char *format_size_aligned(off_t size)
 {
 	static	char *buffer = NULL;
 	char	units;
@@ -407,9 +308,7 @@ char *format_size_aligned(off_t size)
 	else
 		units = ' ';
 
-	if (buffer)
-		g_free(buffer);
-
+	g_free(buffer);
 	buffer = g_strdup_printf("%4" SIZE_FMT "%c", size, units);
 	
 	return buffer;
@@ -420,10 +319,10 @@ char *format_size_aligned(off_t size)
  * unsigned long isn't wide enough on all platforms and we must be able to
  * sum sizes above 4 GB.
  */
-gchar *format_double_size(double size)
+const gchar *format_double_size(double size)
 {
 	static gchar	*buf = NULL;
-	char		*units;
+	const char	*units;
 
 	if (size >= PRETTY_SIZE_LIMIT)
 	{
@@ -437,10 +336,10 @@ gchar *format_double_size(double size)
 			{
 				size += 1023;
 				size /= 1024;
-				units = "Gb";
+				units = "G";
 			}
 			else
-				units = "Mb";
+				units = "M";
 		}
 		else
 			units = "K";
@@ -451,20 +350,12 @@ gchar *format_double_size(double size)
 	else
 		units = _("byte");
 
-	if (buf)
-		g_free(buf);
-	buf = g_strdup_printf("%.0f %s", size, units);
+	g_free(buf);
+	buf = g_strdup_printf("%.0f %s", floor(size), units);
 
 	return buf;
 }
 
-/* Ensure that str ends with a newline (or is empty) */
-static void newline(GString *str)
-{
-	if (str->len && str->str[str->len - 1] != '\n')
-		g_string_append_c(str, '\n');
-}
-
 /* Fork and exec argv. Wait and return the child's exit status.
  * -1 if spawn fails.
  * Returns the error string from the command if any, or NULL on success.
@@ -472,87 +363,42 @@ static void newline(GString *str)
  * a suitable message is created.
  * g_free() the result.
  */
-char *fork_exec_wait(char **argv)
+char *fork_exec_wait(const char **argv)
 {
-	pid_t	child;
-	int	status = -1;
-	GString *errors;
-	char	buffer[257];
-	int	from_stderr;
+	int	status;
+	gchar	*errors = NULL;
+	GError	*error = NULL;
 
-	errors = g_string_new(NULL);
-
-	child = spawn_full(argv, NULL, &from_stderr);
-
-	if (!child)
+	if (!g_spawn_sync(NULL, (char **) argv, NULL,
+		     G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL,
+		     NULL, NULL,
+		     NULL, &errors, &status, &error))
 	{
-		newline(errors);
-		g_string_append(errors, "fork: ");
-		g_string_append(errors, g_strerror(errno));
-		goto out;
+		char *msg;
+
+		msg = g_strdup(error->message);
+		g_error_free(error);
+		return msg;
 	}
 
-	while (from_stderr != -1)
-	{
-		int got;
+	if (errors && !*errors)
+		null_g_free(&errors);
 
-		got = read(from_stderr, buffer, sizeof(buffer) - 1);
-		if (got < 0)
-		{
-			newline(errors);
-			g_string_append(errors, "read: ");
-			g_string_append(errors, g_strerror(errno));
-		}
-		if (got <= 0)
-			break;
-		buffer[got] = '\0';
-		g_string_append(errors, buffer);
+	if (!WIFEXITED(status))
+	{
+		if (!errors)
+			errors = g_strdup("(Subprocess crashed?)");
+	}
+	else if (WEXITSTATUS(status))
+	{
+		if (!errors)
+			errors = g_strdup(_("ERROR"));
 	}
 
-	while (1)
-	{
-		if (waitpid(child, &status, 0) == -1)
-		{
-			if (errno != EINTR)
-			{
-				newline(errors);
-				g_string_append(errors, "waitpid: ");
-				g_string_append(errors, g_strerror(errno));
-				break;
-			}
-		}
-		else
-		{
-			if (!WIFEXITED(status))
-			{
-				newline(errors);
-				g_string_append(errors, "(crashed?)");
-			}
-			else if (WEXITSTATUS(status))
-			{
-				newline(errors);
-				if (!errors->len)
-					g_string_append(errors, "ERROR");
-			}
-			break;
-		}
-	}
+	if (errors)
+		g_strstrip(errors);
 
-out:
-	if (from_stderr != -1)
-		close(from_stderr);
-
-	if (errors->len && errors->str[errors->len - 1] == '\n')
-		g_string_truncate(errors, errors->len - 1);
-
-	if (errors->len)
-	{
-		char *retval = errors->str;
-		g_string_free(errors, FALSE);
-		return retval;
-	}
-
-	return NULL;
+	return errors;
 }
 
 /* If a file has this UID and GID, which permissions apply to us?
@@ -580,9 +426,13 @@ gint applicable(uid_t uid, gid_t gid)
 /* Converts a file's mode to a string. Result is a pointer
  * to a static buffer, valid until the next call.
  */
-char *pretty_permissions(mode_t m)
+const char *pretty_permissions(mode_t m)
 {
-	static char buffer[] = "rwx,rwx,rwx/UGT";
+	static char buffer[] = "rwx,rwx,rwx/UG"
+#ifdef S_ISVTX
+	     "T"
+#endif
+	     ;
 
 	buffer[0]  = m & S_IRUSR ? 'r' : '-';
 	buffer[1]  = m & S_IWUSR ? 'w' : '-';
@@ -600,9 +450,6 @@ char *pretty_permissions(mode_t m)
 	buffer[13] = m & S_ISGID ? 'G' : '-';
 #ifdef S_ISVTX
         buffer[14] = m & S_ISVTX ? 'T' : '-';
-        buffer[15] = 0;
-#else
-        buffer[14] = 0;
 #endif
 
 	return buffer;
@@ -619,34 +466,36 @@ static gboolean is_local_address(char *address)
 }
 
 /* Convert a URI to a local pathname (or NULL if it isn't local).
- * The returned pointer points inside the input string.
+ * The returned pointer needs to be passed to g_free() when done (if not NULL).
+ * THIS IS A CHANGE. The return path has been processed by unescape_uri().
  * Possible formats:
  *	/path
  *	///path
  *	//host/path
  *	file://host/path
  */
-char *get_local_path(char *uri)
+char *get_local_path(const char *uri)
 {
 	if (*uri == '/')
 	{
 		char    *path, *uri_host;
 
+		/* Just a local path - no host part */
 		if (uri[1] != '/')
-			return uri;	/* Just a local path - no host part */
-
+			return unescape_uri(uri);
+		                    
 		path = strchr(uri + 2, '/');
 		if (!path)
 			return NULL;	    /* //something */
 
 		if (path - uri == 2)
-			return path;	/* ///path */
+			return unescape_uri(path);	/* ///path */
 		
 		uri_host = g_strndup(uri + 2, path - uri - 2);
 		if (is_local_address(uri_host))
 		{
 			g_free(uri_host);
-			return path;	/* //myhost/path */
+			return unescape_uri(path);	/* //myhost/path */
 		}
 		g_free(uri_host);
 
@@ -681,106 +530,35 @@ void set_blocking(int fd, gboolean blocking)
 		g_warning("fcntl() failed: %s\n", g_strerror(errno));
 }
 
-/* Format this time nicely. The result is a pointer to a static buffer,
- * valid until the next call.
+/* Format this time nicely.
+ * g_free() the result.
  */
-char *pretty_time(time_t *time)
+char *pretty_time(const time_t *time)
 {
-        static char time_buf[32];
+        char time_buf[32];
 
         if (strftime(time_buf, sizeof(time_buf),
 			TIME_FORMAT, localtime(time)) == 0)
 		time_buf[0]= 0;
 
-	return time_buf;
+	return to_utf8(time_buf);
 }
 
 #ifndef O_NOFOLLOW
 #  define O_NOFOLLOW 0x0
 #endif
 
-#ifdef HAVE_LIBVFS
-/* Copy data from 'read_fd' FD to 'write_fd' FD until EOF or error.
- * Returns 0 on success, -1 on error (and sets errno).
- */
-static int copy_fd(int read_fd, int write_fd)
-{
-	char	buffer[4096];
-	int	got;
-	
-	while ((got = mc_read(read_fd, buffer, sizeof(buffer))) > 0)
-	{
-		int sent = 0;
-		
-		while (sent < got)
-		{
-			int	c;
-			
-			c = mc_write(write_fd, buffer + sent, got - sent);
-			if (c < 0)
-				return -1;
-			sent += c;
-		}
-	}
-
-	return got;
-}
-#endif
-
 /* 'from' and 'to' are complete pathnames of files (not dirs or symlinks).
  * This spawns 'cp' to do the copy if lstat() succeeds, otherwise we
  * do the copy manually using vfs.
  *
  * Returns an error string, or NULL on success. g_free() the result.
+ *
+ * XXX: This was only used for libvfs...
  */
-guchar *copy_file(guchar *from, guchar *to)
+guchar *copy_file(const guchar *from, const guchar *to)
 {
-	char	*argv[] = {"cp", "-pRf", NULL, NULL, NULL};
-
-#ifdef HAVE_LIBVFS
-	struct	stat	info;
-
-	if (lstat(from, &info))
-	{
-		int	read_fd = -1;
-		int	write_fd = -1;
-		int	error = 0;
-
-		if (mc_lstat(from, &info))
-			return g_strdup(g_strerror(errno));
-
-		/* Regular lstat() can't find it, but mc_lstat() can,
-		 * so try reading it with VFS.
-		 */
-
-		if (!S_ISREG(info.st_mode))
-			goto err;
-
-		read_fd = mc_open(from, O_RDONLY);
-		if (read_fd == -1)
-			goto err;
-		write_fd = mc_open(to,
-				O_NOFOLLOW | O_WRONLY | O_CREAT | O_TRUNC,
-				info.st_mode & 0777);
-		if (write_fd == -1)
-			goto err;
-
-		if (copy_fd(read_fd, write_fd))
-			goto err;
-
-		/* (yes, the single | is right) */
-		if (mc_close(read_fd) | mc_close(write_fd))
-			return g_strdup(g_strerror(errno));
-		return NULL;
-err:
-		error = errno;
-		if (read_fd != -1)
-			mc_close(read_fd);
-		if (write_fd != -1)
-			mc_close(write_fd);
-		return g_strdup(error ? g_strerror(error) : _("Copy error"));
-	}
-#endif
+	const char *argv[] = {"cp", "-pRf", NULL, NULL, NULL};
 
 	argv[2] = from;
 	argv[3] = to;
@@ -792,7 +570,7 @@ err:
  * into a shell command.
  * Eg: 'My Dir?' becomes 'My\ Dir\?'. g_free() the result.
  */
-guchar *shell_escape(guchar *word)
+guchar *shell_escape(const guchar *word)
 {
 	GString	*tmp;
 	guchar	*retval;
@@ -816,9 +594,10 @@ guchar *shell_escape(guchar *word)
  * (or the two are the same item/directory).
  * FALSE if parent doesn't exist.
  */
-gboolean is_sub_dir(char *sub, char *parent)
+gboolean is_sub_dir(const char *sub_obj, const char *parent)
 {
 	struct stat parent_info;
+	char *sub;
 
 	if (mc_lstat(parent, &parent_info))
 		return FALSE;		/* Parent doesn't exist */
@@ -831,7 +610,7 @@ gboolean is_sub_dir(char *sub, char *parent)
 	 * When checking if an icon depends on a file (parent), use realpath on
 	 * sub (the icon) too.
 	 */
-	sub = pathdup(sub);
+	sub = pathdup(sub_obj);
 	
 	while (1)
 	{
@@ -870,7 +649,7 @@ gboolean is_sub_dir(char *sub, char *parent)
 /* True if the string 'list' contains 'item'.
  * Eg ("close", "close, help") -> TRUE
  */
-gboolean in_list(guchar *item, guchar *list)
+gboolean in_list(const guchar *item, const guchar *list)
 {
 	int	len;
 
@@ -878,12 +657,13 @@ gboolean in_list(guchar *item, guchar *list)
 	
 	while (*list)
 	{
-		if (strncmp(item, list, len) == 0 && !isalpha(list[len]))
+		if (strncmp(item, list, len) == 0 &&
+		    !g_ascii_isalpha(list[len]))
 			return TRUE;
 		list = strchr(list, ',');
 		if (!list)
 			return FALSE;
-		while (isspace(*++list))
+		while (g_ascii_isspace(*++list))
 			;
 	}
 
@@ -898,7 +678,7 @@ gboolean in_list(guchar *item, guchar *list)
  *
  * The array and the strings in it must be freed after use.
  */
-GPtrArray *split_path(guchar *path)
+GPtrArray *split_path(const guchar *path)
 {
 	GPtrArray *array;
 	guchar	*slash;
@@ -931,7 +711,7 @@ GPtrArray *split_path(guchar *path)
 /* Return the shortest path from 'from' to 'to'.
  * Eg: get_relative_path("/a/b/c", "a/d/e") -> "../d/e"
  */
-guchar *get_relative_path(guchar *from, guchar *to)
+guchar *get_relative_path(const guchar *from, const guchar *to)
 {
 	GString  *path;
 	guchar	 *retval;
@@ -997,30 +777,25 @@ guchar *get_relative_path(guchar *from, guchar *to)
  */
 int text_to_boolean(const char *text, int defvalue)
 {
-	if(g_strcasecmp(text, "true")==0)
+	if (g_strcasecmp(text, "true")==0)
 	        return TRUE;
-	else if(g_strcasecmp(text, "false")==0)
+	else if (g_strcasecmp(text, "false")==0)
 	        return FALSE;
-	else if(g_strcasecmp(text, "yes")==0)
+	else if (g_strcasecmp(text, "yes")==0)
 	        return TRUE;
-	else if(g_strcasecmp(text, "no")==0)
+	else if (g_strcasecmp(text, "no")==0)
 	        return FALSE;
-	else if(isdigit(text[0]))
+	else if (g_ascii_isdigit(text[0]))
 	        return !!atoi(text);
 
 	return defvalue;
 }
 
-void set_to_null(gpointer *data)
-{
-	*data = NULL;
-}
-
 /* Return the pathname that this symlink points to.
  * NULL on error (not a symlink, path too long) and errno set.
  * g_free() the result.
  */
-char *readlink_dup(char *source)
+char *readlink_dup(const char *source)
 {
 	char	path[MAXPATHLEN + 1];
 	int	got;
@@ -1032,7 +807,6 @@ char *readlink_dup(char *source)
 	return g_strndup(path, got);
 }
 
-#if defined(GTK2) || defined(THUMBS_USE_LIBPNG)
 /*
  * This code implements the MD5 message-digest algorithm.
  * The algorithm is due to Ron Rivest. The original code was
@@ -1270,7 +1044,7 @@ static void MD5Transform(guint32 buf[4], guint32 const in[16])
 
 # endif /* ASM_MD5 */
 
-char *md5_hash(char *message)
+char *md5_hash(const char *message)
 {
 	MD5Context ctx;
 
@@ -1278,12 +1052,41 @@ char *md5_hash(char *message)
 	MD5Update(&ctx, message, strlen(message));
 	return MD5Final(&ctx);
 }
-#endif /* GTK2 or THUMBS_USE_LIBPNG */
+
+/* Convert string 'src' from the current locale to UTF-8 */
+gchar *to_utf8(const gchar *src)
+{
+	gchar *retval;
+
+	if (!src)
+		return NULL;
+
+	retval = g_locale_to_utf8(src, -1, NULL, NULL, NULL);
+	if (!retval)
+		retval = g_convert_with_fallback(src, -1, "utf-8", "iso-8859-1",
+						"?", NULL, NULL, NULL);
+
+	return retval ? retval : g_strdup(src);
+}
+
+/* Ensure string at 'sp' is UTF-8. g_free() and replace by
+ * UTF-8 version if not.
+ */
+void ensure_utf8(gchar **sp)
+{
+	gchar *s = *sp;
+
+	if (!g_utf8_validate(s, -1, NULL))
+	{
+		*sp = to_utf8(s);
+		g_free(s);
+	}
+}
 
 /* Removes trailing / chars and converts a leading '~/' (if any) to
  * the user's home dir. g_free() the result.
  */
-gchar *icon_convert_path(gchar *path)
+gchar *expand_path(const gchar *path)
 {
 	guchar		*retval;
 	int		path_len;
@@ -1307,68 +1110,294 @@ gchar *icon_convert_path(gchar *path)
 	return retval;
 }
 
-/* Convert string 'src' from the current locale to UTF-8 */
-gchar *to_utf8(gchar *src)
+/* g_free() every element in the list, then free the list itself and
+ * NULL the pointer to the list.
+ */
+void destroy_glist(GList **list)
 {
-	gchar *retval;
-
-	if (!src)
-		return NULL;
-
-	retval = g_locale_to_utf8(src, -1, NULL, NULL, NULL);
-
-	return retval ? retval : g_strdup(src);
+	GList *l = *list;
+	g_list_foreach(l, (GFunc) g_free, NULL);
+	g_list_free(l);
+	*list = NULL;
 }
 
-/* Convert string 'src' to the current locale from UTF-8 */
-gchar *from_utf8(gchar *src)
+void null_g_free(gpointer p)
 {
-	gchar *retval;
-	
-	if (!src)
-		return NULL;
-
-	retval = g_locale_from_utf8(src, -1, NULL, NULL, NULL);
-
-	return retval ? retval : g_strdup(src);
+	g_free(*(gpointer *)p);
+	*(gpointer *)p = NULL;
 }
 
-/****************************************************************
- *                      INTERNAL FUNCTIONS                      *
- ****************************************************************/
+typedef struct _CollatePart CollatePart;
 
-static XMLwrapper *xml_load(char *pathname, gpointer data)
+struct _CollateKey {
+	CollatePart *parts;
+	gboolean caps;
+};
+
+struct _CollatePart {
+	guchar *text;	/* NULL => end of list */
+	long number;
+};
+
+/* Break 'name' (a UTF-8 string) down into a list of (text, number) pairs.
+ * The text parts processed for collating. This allows any two names to be
+ * quickly compared later for intelligent sorting (comparing names is
+ * speed-critical).
+ */
+CollateKey *collate_key_new(const guchar *name)
 {
-	xmlDocPtr doc;
-	XMLwrapper *xml_data = NULL;
+	const guchar *i;
+	guchar *to_free = NULL;
+	GArray *array;
+	CollatePart new;
+	CollateKey *retval;
+	char *tmp;
 
-	doc = xmlParseFile(pathname);
-	if (!doc)
-		return NULL;	/* Bad XML */
+	g_return_val_if_fail(name != NULL, NULL);
 
-	xml_data = g_new(XMLwrapper, 1);
-	xml_data->ref = 1;
-	xml_data->doc = doc;
+	array = g_array_new(FALSE, FALSE, sizeof(CollatePart));
 
-	return xml_data;
-}
-
-static void xml_ref(XMLwrapper *doc, gpointer data)
-{
-	if (doc)
-		doc->ref++;
-}
-
-static void xml_unref(XMLwrapper *doc, gpointer data)
-{
-	if (doc && --doc->ref == 0)
+	/* Ensure valid UTF-8 */
+	if (!g_utf8_validate(name, -1, NULL))
 	{
-		xmlFreeDoc(doc->doc);
-		g_free(doc);
+		to_free = to_utf8(name);
+		name = to_free;
+	}
+
+	retval = g_new(CollateKey, 1);
+	retval->caps = g_unichar_isupper(g_utf8_get_char(name));
+
+	for (i = name; *i; i = g_utf8_next_char(i))
+	{
+		/* We're in a (possibly blank) text section starting at 'name'.
+		 * Find the end of it (the next digit, or end of string).
+		 */
+		if (g_unichar_isdigit(g_utf8_get_char(i)))
+		{
+			char *endp;
+			
+			/* i -> first digit character */
+			tmp = g_utf8_strdown(name, i - name);
+			new.text = g_utf8_collate_key(tmp, -1);
+			g_free(tmp);
+			new.number = strtol(i, &endp, 10);
+
+			g_array_append_val(array, new);
+
+			g_return_val_if_fail(endp > (char *) i, NULL);
+
+			name = endp;
+			i = name - 1;
+		}
+	}
+
+	tmp = g_utf8_strdown(name, i - name);
+	new.text = g_utf8_collate_key(tmp, -1);
+	g_free(tmp);
+	new.number = -1;
+	g_array_append_val(array, new);
+
+	new.text = NULL;
+	g_array_append_val(array, new);
+
+	retval->parts = (CollatePart *) array->data;
+	g_array_free(array, FALSE);
+
+	if (to_free)
+		g_free(to_free);	/* Only taken for invalid UTF-8 */
+
+	return retval;
+}
+
+void collate_key_free(CollateKey *key)
+{
+	CollatePart *part;
+
+	for (part = key->parts; part->text; part++)
+		g_free(part->text);
+	g_free(key->parts);
+	g_free(key);
+}
+
+int collate_key_cmp(const CollateKey *key1, const CollateKey *key2,
+		    gboolean caps_first)
+{
+	CollatePart *n1 = key1->parts;
+	CollatePart *n2 = key2->parts;
+	int r;
+
+	if (caps_first)
+	{
+		if (key1->caps && !key2->caps)
+			return -1;
+		else if (key2->caps && !key1->caps)
+			return 1;
+	}
+
+	while (1)
+	{
+		if (!n1->text)
+			return n2->text ? -1 : 0;
+		if (!n2->text)
+			return 1;
+		r = strcmp(n1->text, n2->text);
+		if (r)
+			return r;
+
+		if (n1->number < n2->number)
+			return -1;
+		if (n1->number > n2->number)
+			return 1;
+
+		n1++;
+		n2++;
 	}
 }
 
-static int xml_getref(XMLwrapper *doc, gpointer data)
+/* Returns TRUE if the object exists, FALSE if it doesn't.
+ * For symlinks, the file pointed to must exist.
+ */
+gboolean file_exists(const char *path)
 {
-	return doc ? doc->ref : 0;
+	struct stat info;
+
+	return !mc_stat(path, &info);
+}
+
+/* Escape path for future use in URI */
+gchar *escape_uri_path(const char *path)
+{
+	const char *safe = "-_./"; /* Plus isalnum() */
+	const guchar *s;
+	gchar *ans;
+	GString *str;
+
+	str = g_string_sized_new(strlen(path));
+
+	for (s = path; *s; s++)
+	{
+		if (!g_ascii_isalnum(*s) && !strchr(safe, *s))
+			g_string_append_printf(str, "%%%02x", *s);
+		else
+			str = g_string_append_c(str, *s);
+	}
+
+	ans = str->str;
+	g_string_free(str, FALSE);
+
+	return ans;
+}
+
+gchar *encode_path_as_uri(const guchar *path)
+{
+	gchar *tpath = escape_uri_path(path);
+	gchar *uri;
+
+	uri = g_strconcat("file://", our_host_name_for_dnd(), tpath, NULL);
+	g_free(tpath);
+
+	return uri;
+}
+
+gchar *unescape_uri(const char *uri)
+{
+	const gchar *s;
+	gchar *d;
+	gchar *tmp;
+	
+	tmp = g_malloc(strlen(uri) + 1);
+	for (s = uri, d = tmp; *s; s++, d++)
+	{
+		/*printf("%s\n", s);*/
+		if (*s == '%' && g_ascii_isxdigit(s[1]) &&
+				 g_ascii_isxdigit(s[2]))
+		{
+			int c;
+			char buf[3];
+			buf[0] = s[1];
+			buf[1] = s[2];
+			buf[2] = 0;
+			c = (int) strtol(buf, NULL, 16);
+			*d = c;
+			s += 2;
+		}
+		else
+			*d = *s;
+	}
+	*d = '\0';
+
+	return tmp;
+}
+
+/* Used as the sort function for sorting GPtrArrays */
+gint strcmp2(gconstpointer a, gconstpointer b)
+{
+	const char *aa = *(char **) a;
+	const char *bb = *(char **) b;
+
+	return g_strcasecmp(aa, bb);
+}
+
+/* Returns an array listing all the names in the directory 'path'.
+ * The array is sorted.
+ * '.' and '..' are skipped.
+ * On error, the error is reported with g_warning and NULL is returned.
+ */
+GPtrArray *list_dir(const guchar *path)
+{
+	GDir *dir;
+	GError *error = NULL;
+	const char *leaf;
+	GPtrArray *names;
+	
+	dir = g_dir_open(path, 0, &error);
+	if (error)
+	{
+		g_warning("Can't list directory:\n%s", error->message);
+		g_error_free(error);
+		return NULL;
+	}
+
+	names = g_ptr_array_new();
+
+	while ((leaf = g_dir_read_name(dir))) {
+		if (leaf[0] != '.')
+			g_ptr_array_add(names, g_strdup(leaf));
+	}
+
+	g_dir_close(dir);
+
+	g_ptr_array_sort(names, strcmp2);
+
+	return names;
+}
+
+int stat_with_timeout(const char *path, struct stat *info)
+{
+	int status;
+	pid_t child;
+	gboolean retval;
+
+	child = fork();
+	if (child < 0)
+	{
+		g_warning("stat_with_timeout: fork(): %s", g_strerror(errno));
+		return -1;
+	}
+
+	if (child == 0)
+	{
+		/* Child */
+		alarm(3);
+		_exit(mc_stat(path, info) ? 1 : 0);
+	}
+
+	waitpid(child, &status, 0);
+
+	if (status == 0)
+		retval = mc_stat(path, info);
+	else
+		retval = -1;
+
+	return retval;
 }
